/*--------------------------------------------------------------------------
    Pawny 0.1.6, chess engine (source code).
    Copyright (C) 2009 - 2010 by Mincho Georgiev.
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    contact: pawnychess@gmail.com 
    web: http://www.pawny.netii.net/
----------------------------------------------------------------------------*/

#include "data.h"


static void move_capture(move_t ms[],int *pcount,int from, int to)
{	
	ms[*pcount].p = 0;
	ms[*pcount].score = 0;
	ms[*pcount].from = (sq_t)from;
	ms[*pcount].to = (sq_t)to;
	ms[*pcount].type = CAP;
	*pcount += 1;
}

static void move_ep(move_t ms[],int *pcount, int from, int to)
{	
	ms[*pcount].p = 0;
	ms[*pcount].score = 0;
	ms[*pcount].from = (sq_t)from;
	ms[*pcount].to = (sq_t)to;
	ms[*pcount].type = EP | CAP;
	*pcount += 1;
}

static void move_push(move_t ms[],int *pcount, int from, int to,int castle_flag)
{
	ms[*pcount].p = 0;
	ms[*pcount].score = 0;
	ms[*pcount].from = (sq_t)from;
	ms[*pcount].to = (sq_t)to;
	if(castle_flag) ms[*pcount].type = (uint8)(CASTLE | castle_flag);
	*pcount += 1;
}

static void move_promote(move_t ms[],int *pcount,int from, int to)
{
	int i;
	for(i = QUEEN; i >= KNIGHT; i--)
	{	ms[*pcount].p = 0;
		ms[*pcount].score = 0;
		ms[*pcount].from = (sq_t)from;
		ms[*pcount].to = (sq_t)to;
		ms[*pcount].promoted = (uint8)Coloured(i);
		ms[*pcount].type = PROM;
		if(!IsEmpty(to)) ms[*pcount].type |= CAP;
		*pcount += 1;
	}
}

int move_gen(move_t ms[])
{
	const int *dir;
	int rank;
	int from,to;
	int color = board->side;
	int opcolor = board->xside;
	int count = 0;
	
	//pawn moves:
	for(from = PLS(Coloured(PAWN)); from <= H8; from = PLN(from))
	{
		if(color == WHITE)
		{
			rank = calc_rank(from);
			to = from + 15;
			if(PieceColor(to) == opcolor){
				if(rank == RANK_7) move_promote(&ms[0],&count, from, to);
				else move_capture(&ms[0],&count, from, to);
			}
			to = from + 17;
			if(PieceColor(to) == opcolor){
				if(rank == RANK_7) move_promote(&ms[0],&count, from, to);
				else move_capture(&ms[0],&count, from, to);
			}
			to = from + 16;
			if(IsEmpty(to)){	
				if(rank == RANK_7) move_promote(&ms[0],&count, from, to);
				else move_push(&ms[0],&count, from, to, 0);
				to += 16;
				if((rank == RANK_2) && IsEmpty(to)) move_push(&ms[0],&count, from, to, 0);
			}
		}
		else
		{
			rank = calc_rank(from);
			to = from - 15;
			if(PieceColor(to) == opcolor){
				if(rank == RANK_2) move_promote(&ms[0],&count,  from, to);
				else move_capture(&ms[0],&count, from, to);
			}
			to = from - 17;
			if(PieceColor(to) == opcolor){
				if(rank == RANK_2) move_promote(&ms[0],&count, from, to);
				else move_capture(&ms[0],&count, from, to);
			}
			to = from - 16;
			if (IsEmpty(to)){	
				if(rank == RANK_2) move_promote(&ms[0],&count, from, to);
				else move_push(&ms[0],&count, from, to, 0);
				to -= 16;
				if((rank == RANK_7) && IsEmpty(to)) move_push(&ms[0],&count, from, to, 0);
			}
		}
	}
	//queen moves:
	for(from = PLS(Coloured(QUEEN)); from <= H8; from = PLN(from))
	{	for(dir = dir_vect[QUEEN]; *dir != 0; dir++)
		{	for(to = from + *dir; IsEmpty(to); to = to + *dir)
				move_push(&ms[0],&count, from, to, 0);
			if(PieceColor(to) == opcolor)
				move_capture(&ms[0],&count, from, to);
		}
	}
	//rooks:
	for(from = PLS(Coloured(ROOK)); from <= H8; from = PLN(from))
	{	for(dir = dir_vect[ROOK]; *dir != 0; dir++)
		{	for(to = from + *dir; IsEmpty(to); to = to + *dir)
				move_push(&ms[0],&count, from, to, 0);
			if(PieceColor(to) == opcolor)
				move_capture(&ms[0],&count, from, to);
		}
	}
	//bishops
	for(from = PLS(Coloured(BISHOP)); from <= H8; from = PLN(from))
	{	for(dir = dir_vect[BISHOP]; *dir != 0; dir++)
		{	for(to = from + *dir; IsEmpty(to); to = to + *dir)
				move_push(&ms[0],&count, from, to, 0);
			if(PieceColor(to) == opcolor)
				move_capture(&ms[0],&count, from, to);
		}
	}
	//knights
	for(from = PLS(Coloured(KNIGHT)); from <= H8; from = PLN(from))
	{	for(dir = dir_vect[KNIGHT]; *dir != 0; dir++)
		{	to = from + *dir;
			if(IsEmpty(to)) move_push(&ms[0],&count, from, to, 0);
			if(PieceColor(to) == opcolor)
				move_capture(&ms[0],&count, from, to);
		}
	}
	//king:
	from = PLS(Coloured(KING));
	for(dir = dir_vect[KING]; *dir != 0; dir++)
	{	to = from + *dir;
		if(IsEmpty(to)) move_push(&ms[0],&count, from, to, 0);
		if(PieceColor(to) == opcolor)
			move_capture(&ms[0],&count, from, to);
	}
	//castling
	if(color == WHITE)
	{	if(board->castle & WHITE_OO)
			if(IsEmpty(F1) && IsEmpty(G1) 
			&& (PieceType(H1) == WHITE_ROOK))
				move_push(&ms[0],&count, E1, G1, WHITE_OO);

		if(board->castle & WHITE_OOO)
			if(IsEmpty(D1) && IsEmpty(C1) 
			&& IsEmpty(B1) && (PieceType(A1) == WHITE_ROOK))
				move_push(&ms[0],&count, E1, C1, WHITE_OOO);
	}
	else//black to move:
	{	if(board->castle & BLACK_OO)
			if(IsEmpty(F8) && IsEmpty(G8) 
			&& (PieceType(H8) == BLACK_ROOK))
				move_push(&ms[0],&count, E8, G8, BLACK_OO);

		if(board->castle & BLACK_OOO) 
			if(IsEmpty(D8) && IsEmpty(C8) 
			&& IsEmpty(B8) && (PieceType(A8) == BLACK_ROOK))
				move_push(&ms[0],&count, E8, C8, BLACK_OOO);
	}
	//en-passant
	if(board->en_passant)
	{	from = board->en_passant + dir_vect[OpColoured(PAWN)][0];
		to = from + dir_vect[Coloured(PAWN)][0];
		if(PieceType(from) == Coloured(PAWN))
			move_ep(&ms[0],&count, from, to);

		from = board->en_passant + dir_vect[OpColoured(PAWN)][1];
		to = from + dir_vect[Coloured(PAWN)][1];
		if(PieceType(from) == Coloured(PAWN))
			move_ep(&ms[0],&count, from, to);
	}
	return count;
}


int move_gen_caps(move_t ms[])
{
	const int *dir;
	int from,to,rank;
	int color = board->side;
	int opcolor = board->xside;
	int count = 0;
	
	//pawn moves:
	for(from = PLS(Coloured(PAWN)); from <= H8; from = PLN(from))
	{
		if(color == WHITE)
		{
			rank = calc_rank(from);
			to = from + 15;
			if(PieceColor(to) == opcolor){
				if(rank == RANK_7) move_promote(&ms[0],&count, from, to);
				else move_capture(&ms[0],&count, from, to);
			}
			to = from + 17;
			if(PieceColor(to) == opcolor){
				if(rank == RANK_7) move_promote(&ms[0],&count, from, to);
				else move_capture(&ms[0],&count, from, to);
			}
			to = from + 16;
			if(IsEmpty(to)){	
				if(rank == RANK_7) move_promote(&ms[0],&count, from, to);
			}
		}
		else
		{
			rank = calc_rank(from);
			to = from - 15;
			if(PieceColor(to) == opcolor){
				if(rank == RANK_2) move_promote(&ms[0],&count,  from, to);
				else move_capture(&ms[0],&count, from, to);
			}
			to = from - 17;
			if(PieceColor(to) == opcolor){
				if(rank == RANK_2) move_promote(&ms[0],&count, from, to);
				else move_capture(&ms[0],&count, from, to);
			}
			to = from - 16;
			if (IsEmpty(to)){	
				if(rank == RANK_2) move_promote(&ms[0],&count, from, to);
			}
		}
	}
	//queen moves:
	for(from = PLS(Coloured(QUEEN)); from <= H8; from = PLN(from))
	{	for(dir = dir_vect[QUEEN]; *dir != 0; dir++)
		{	for(to = from + *dir; IsEmpty(to); to = to + *dir);
			if(PieceColor(to) == opcolor)
				move_capture(&ms[0],&count, from, to);
		}
	}
	//rooks:
	for(from = PLS(Coloured(ROOK)); from <= H8; from = PLN(from))
	{	for(dir = dir_vect[ROOK]; *dir != 0; dir++)
		{	for(to = from + *dir; IsEmpty(to); to = to + *dir);
			if(PieceColor(to) == opcolor)
				move_capture(&ms[0],&count, from, to);
		}
	}
	//bishops
	for(from = PLS(Coloured(BISHOP)); from <= H8; from = PLN(from))
	{	for(dir = dir_vect[BISHOP]; *dir != 0; dir++)
		{	for(to = from + *dir; IsEmpty(to); to = to + *dir);
			if(PieceColor(to) == opcolor)
				move_capture(&ms[0],&count, from, to);
		}
	}
	//knights
	for(from = PLS(Coloured(KNIGHT)); from <= H8; from = PLN(from))
	{	for(dir = dir_vect[KNIGHT]; *dir != 0; dir++)
		{	to = from + *dir;
			if(PieceColor(to) == opcolor)
				move_capture(&ms[0],&count, from, to);
		}
	}
	//king:
	from = PLS(Coloured(KING));
	for(dir = dir_vect[KING]; *dir != 0; dir++)
	{	to = from + *dir;
		if(PieceColor(to) == opcolor)
			move_capture(&ms[0],&count, from, to);
	}
	//en-passant
	if(board->en_passant)
	{	from = board->en_passant + dir_vect[OpColoured(PAWN)][0];
		to = from + dir_vect[Coloured(PAWN)][0];
		if(PieceType(from) == Coloured(PAWN))
			move_ep(&ms[0],&count, from, to);

		from = board->en_passant + dir_vect[OpColoured(PAWN)][1];
		to = from + dir_vect[Coloured(PAWN)][1];
		if(PieceType(from) == Coloured(PAWN))
			move_ep(&ms[0],&count, from, to);
	}
	return count;

}

int move_gen_legal(move_t ms[])
{///returns legal moves only, 
///temporary (and very expensive) solution for the single reply ext.!
	int i,count;
	int j = 0;
	move_t temp_ms[MOVE_STACK];
	
		count = move_gen(&temp_ms[0]);
		for(i = 0; i < count; i++)
		{
			if(!move_make(temp_ms[i])) continue;
			move_undo();
			ms[j] = temp_ms[i];
			j++;
		}
		return j;
}

