/*--------------------------------------------------------------------------
    Pawny 0.1.6, chess engine (source code).
    Copyright (C) 2009 - 2010 by Mincho Georgiev.
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    contact: pawnychess@gmail.com 
    web: http://www.pawny.netii.net/
----------------------------------------------------------------------------*/

#include "data.h"

#define HINDEX ((board->hash % opt->tt_numentry))


void tt_free() { if(tt) free(tt); }
void tt_clear(){ memset(tt, 0, (size_t)(opt->tt_size));}

int tt_init(int MB)
{///min size - 8MB, max - 1024:
	tt_free();
	if(MB < 8) MB = 8; if(MB > 1024) MB = 1024;
	opt->tt_size = (MB*1024*1024);
	opt->tt_numentry = opt->tt_size / sizeof(tt_t);	
	tt = (tt_t *)malloc((size_t)opt->tt_size);
	if(!tt) return false;
	memset(tt, 0,(size_t)(opt->tt_size));
	si->num_hashrec = 0;
	return true;
}

uint8 tt_probe(int depth, int *value, uint32 *hash_move, uint8 *null_ok)
{
	tt_t *p = &tt[HINDEX];
	if(p->key == board->hash)
	{	*hash_move = p->hash_move;
		*null_ok = (p->flags >> 4);
		if(p->depth >= depth)
		{	*value = p->value;
			return (p->flags & 0x0F);
		}
	}
	return TT_UNKNOWN;
}

void tt_save(int depth, int value, uint8 flag, uint32 hash_move, uint8 null_ok)
{///ALWAYS replacement scheme
	
	tt_t *p = &tt[HINDEX];
	
	if(!p->key) si->num_hashrec++;
	si->num_hash_saves++;
	p->key = board->hash;
	p->value = (sint16)value;
	p->depth = (sint8)depth;
	p->flags = (null_ok << 4) | flag;
 	if(hash_move) p->hash_move = hash_move;
}

void tt_check()
{
	if(si->num_hashrec >= opt->tt_numentry)
	{			
		#ifdef LOGGING
		FILE *logfile = fopen("logfile.txt", "at");
		if (!logfile) logfile = fopen("logfile.txt", "wt");
		if (logfile){
			fprintf(logfile, "tt - cleared, num_records = %llu\n",si->num_hashrec);
			fclose(logfile);
		}
		#endif
		tt_clear();
		si->num_hashrec = 0;
	}
}

uint32 tt_get_hashmove()
{///getting the hash move for the current pos.
	tt_t *p = &tt[HINDEX];
	if(p->key == board->hash)
		return p->hash_move;
	return 0;
}

int tt_retrieve_ponder(move_t *m)
{///retrieving the next move from tt
	move_t pvmove;
	pvmove.p = si->pv[0][0].p;
	m->p = 0;
	move_make(pvmove);
	m->p = tt_get_hashmove();
	if(!m->p)
	{	move_undo();
		return false;
	}
	move_undo();
	return true;
}
